/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.guard.file.storage;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Objects;
import org.json.JSONException;
import org.json.JSONObject;
import com.openexchange.exception.OXException;
import com.openexchange.guard.api.GuardApi;
import com.openexchange.guard.api.GuardApis;
import com.openexchange.guard.internal.authentication.GuardAuthenticationToken;

/**
 * {@link GuardEncryptingInputStream} encrypts data using OX Guard
 *
 * @author <a href="mailto:benjamin.gruedelbach@open-xchange.com">Benjamin Gruedelbach</a>
 * @since v2.6.0
 */
public class GuardEncryptingInputStream extends AbstractGuardInputStream {

    private final GuardApi guardApi;
    private final InputStream plainTextData;
    private final boolean armored;
    private final Collection<String> recipients;
    private final GuardAuthenticationToken signingAuthentication;

    /**
     * Initializes a new {@link GuardEncryptingInputStream}.
     *
     * @param guardApi The Guard endpoint to use
     * @param plainTextData The data to encrypt
     * @param recipients A collection of recipients to encrypt the data for
     * @param armored True, to encrypt the data as ASCII-Armored, false otherwise
     * @param signingAuthentication The authentication token for signing the encrypted data, or null to not sign the encrypted data
     */
    public GuardEncryptingInputStream(GuardApi guardApi, InputStream plainTextData, Collection<String> recipients, boolean armored, GuardAuthenticationToken signingAuthentication) {
        this.guardApi = Objects.requireNonNull(guardApi, "guardApi must not be null");
        this.plainTextData = plainTextData;
        this.recipients = Objects.requireNonNull(recipients, "recipients must not be null");
        this.armored = armored;
        this.signingAuthentication = signingAuthentication;

    }

    /**
     * Initializes a new {@link GuardEncryptingInputStream} for encrypting data without signing it.
     *
     * @param guardApi The Guard endpoint to use
     * @param plainTextData The data to encrypt
     * @param recipients A collection of recipients to encrypt the data for
     * @param armored True, to encrypt the data as ASCII-Armored, false otherwise
     */
    public GuardEncryptingInputStream(GuardApi guardApi, InputStream plainTextData, Collection<String> recipients, boolean armored) {
        this(guardApi, plainTextData, recipients, armored, null);
    }

    /**
     * Initializes a new {@link GuardEncryptingInputStream} for encrypting data without signing it.
     *
     * The output will be binary (NOT ASCII-Armored).
     *
     * @param guardApi
     * @param plainTextData
     * @param recipients
     */
    public GuardEncryptingInputStream(GuardApi guardApi, InputStream plainTextData, Collection<String> recipients) {
        this(guardApi, plainTextData, recipients, false, null);
    }

    /**
     * Internal helper method to put a collection to a JSON structure.
     *
     * @param json The json
     * @param key The key
     * @param collection The collection to add
     * @throws IllegalArgumentException due an internal JSON Exception
     */
    private void putCollectionToJsonSave(JSONObject json, String key, Collection<? extends Object> collection) {
        try {
            json.put(key, collection);
        } catch (JSONException e) {
            throw new IllegalArgumentException(e.getMessage());
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see com.openexchange.guard.file.storage.AbstractGuardInputStream#getGuardStream()
     */
    @Override
    protected InputStream getGuardStream() throws OXException {

        JSONObject json = new JSONObject();
        json.putSafe("armored", armored);
        putCollectionToJsonSave(json, "recipients", recipients);
        if (signingAuthentication != null) {
            JSONObject user = new JSONObject();
            user.putSafe("session", signingAuthentication.getGuardSessionId());
            user.putSafe("auth", signingAuthentication.getValue());
            json.putSafe("user", user);
        }

        ByteArrayInputStream jsonStream = new ByteArrayInputStream(json.toString().getBytes());
        LinkedHashMap<String, InputStream> multipartParameters = new LinkedHashMap<>();
        multipartParameters.put("json", jsonStream);
        multipartParameters.put("data", plainTextData);

        return guardApi.processResources(GuardApis.mapFor("action", "encrypt", "respondWithJSON", "true" /* forcing guard to render errors in JSON */), multipartParameters, null);
    }
}
